#ifndef TS_TYPES_H
#define TS_TYPES_H

/* the following type defines a vertex;
 * x, y, and z are the coordinates of the vertex
 */
typedef struct { double x, y, z; } vertex;

/* sides (edges) of all triangles are stored temporarily in an array of aedges first, 
 * internal edges of "triangulated area" will be included twice in the array, while 
 * boundary edges only once
 * 
 *     v1, v2    vertices connected by the edge under consideration,
 *     ies       -1 or -2 for boundary edges, otherwise
 *               a unique identification number of an internal edge assigned
 *               to both its occurences in the array of aedges;
 *               -2 is used to mark the first edge of an isolated triangle
 */
typedef struct { long v1, v2, ies; } aedge;

/* the cycle type codes sequential cycles
 *
 *     e         the representative edge of the sequential cycle
 *     b         an auxiliary 0-1 variable ( 1 = the representative edge 
 *               has already been chosen for this cycle )    
 *     d         0-1 value coding the direction in which the cycle 
 *               meets the representative edge
 */               
typedef struct {
	long e;
	unsigned char b, d;
} cycle;

/* all internal edges are stored in an array of... edges :-)
 *
 *     v1, v2    vertices connected by the internal edge
 *     e[4]      this array encodes neighbouring edges of the given edge,
 *               in fact, if a neighbouring edge is an internal edge, the 
 *               corresponding element of e[] equals the index of this 
 *               neighbouring edge in the array of internal edges; if the 
 *               neighbouring edge is a boundary edge, the corresponding element 
 *               of e[] equals -i - 1, where i is the identification number of 
 *               the vertex of this boundary edge that is not the vertex
 *               of the edge under consideration (the vertices are numbered 
 *               starting from zero)
 *     ie[4]     let E be the edge under consideration, let e[] be the array
 *               of neighbouring edges of E, and let E' be the neighbouring edge 
 *               of E given by e[i]; if E' is a boundary edge, the value of ie[i]
 *               is not defined; if E' is an internal edge, let e'[] be the array of
 *               neighbouring edges of E'; then ie[i] equals the index j such that
 *               e'[j] points to E 
 *     r         0-1 value characterizing the orientation of the edge 
 *
 *     Variables c[2] and f are used in two contexts. Their first use is made during
 *     determining sequential cycles. At this stage, 
 *
 *     c[0]      is used to carry the identification number of the sequence 
 *               of internal edges tested for being a sequential cycle, that has met 
 *               the given edge most recently; if this edge has not been met by 
 *               any such sequence yet, c[0] is -1; this variable helps to detect
 *               invalid, "self-crossing" cycles;
 *     c[1]      is not used;
 *     f         serves for indication that the sequences of internal edges meeting
 *               the given edge have already been tested for being a cycle; it
 *               equals 0 if no of the two sequences have been tested, it is 1
 *               if solely the sequence meeting the edge in direction 0 has been 
 *               tested, it is 2 if solely the sequence meeting the edge in 
 *               direction 1 has been tested, and it is 3 if both the sequences 
 *               have been tested.
 *
 *     The second use of these variables is made during the process of selecting
 *     the representative edges for the sequential cycles. At this stage,
 *
 *     c[i]      is the identification number of the cycle that meets the edge
 *               in direction i; if there is no cycle meeting the edge in 
 *               direction i, c[i] equals -1;
 *     f         equals 4 if the edge under consideration has been selected
 *               as the representative edge of a cycle; otherwise, it is 0
 *               if no sequential cycle meets the edge, it is 1 or 2 if a unique
 *               sequential cycle meets the edge - in direction 0 or 1, 
 *               respectively, and it is 3 if two sequential cycles meet the
 *               edge.
 */
typedef struct {
	long v1, v2;
	long c[2], e[4];
	unsigned char f, ie[4], r;
} edge;

/* the Hopfield network is represented as an array of neurons
 *
 *     s         0-1 value, the state of the neuron
 *     u         0-1 value, an auxiliary variable used for indication that
 *               a certain action has or has not been performed with the neuron
 *     w0        the bias of the neuron
 *     c         a pointer to the array of neighbours of (neurons connected to)
 *               the neuron
 *     w         a pointer to the array of weights assigned to connections
 *               between the neuron and its neighbours
 *     c_size    the number of neighbours of the neuron
 */
typedef struct {
	unsigned char s, u;
	long w0, *c, *w, c_size;
} neuron;

#endif
