#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include "ts_msgs.h"
#include "input.h"
#include "hnet.h"
#include "comp.h"
#include "tstrips.h"

/* procedure test_file tests whether the file given by str exists, and
 * if it does not, creates it
 */
void test_file ( const char *str )
{
	FILE *file;

	file = fopen ( str, "r" );
	
	if ( !file ) {
		file = fopen ( str, "w" );
		if ( !file ) 
			ts_warning ( "test_file", 6 );
		else
			fclose ( file );
	}
	else
		fclose ( file );
}

int main ( int argc, char *argv[] ) 
{  
	long i, j;
	
	/* the number of trials */
	long nt;
	
	/* the initial temperature and the stopping criterion */
	double t_init, stop_crit;
	
	/* the number of the trial with minimum number of triangle strips */
	long tno;

	/* the running time and the number of iterations of the trial with minimum
	 * number of triangle strips */
	double rtime;
	long iter;
	
	/* an indicator of presence of temporary files */ 
	int no_temp_files;
	
	/* the array of the counts of triangle strip lengths */
	long fcnt[10];
	
	/* the length of the shortest and of the longest triangle strip */
	long minlength, maxlength;
	
	/* the length of the shortest and of the longest triangle strip for the current trial */
	long tminlength, tmaxlength;
	
	/* convergence macroscopic time and the number of sequential strips reached
	 * in a single trial, respectively */
	long tt, ps;

	/* current minimum of the numbers of sequential strips */
	long mps;
	
	/* variables used for measuring the time */
	clock_t st_delay;
	time_t start, end;
	double read_time, create_time, comp_time;

	/* sums of the numbers of sequential strips, of running times, and
	 * of convergence macroscopic times, respectively */
	double palf, pcas, ptau;
  
	/* the program must be run with at least two command-line arguments */
	if ( argc < 3 ) {    
		printf ( "\n" );
		printf ( "Usage: %s <input obj file> <output objf file>\n", argv[0] );
		printf ( "  [<number of trials> [<initial temperature> [<stopping criterion>]]]\n" );    
		printf ( "\n" );    
		printf ( "  the optional parameters control the simulated annealing\n" );
		printf ( "  (default values 10 10 0.3)\n" );
		printf ( "  <stopping criterion>: maximum percentage of unstable units\n" ); 
		printf ( "\n" ); 
	}
	else {
		/* test whether a file with the same name as the output file and the file "comp.txt"
		 * exist; if any of these files do not exist, create them */
		test_file ( argv[2] );
		test_file ( "comp.txt" );
		
		/* read the data from the input file and create the array of vertices
		 * and the array of internal edges */
		printf ( "\n" );   
		printf ( "Reading \"%s\"... ", argv[1] );
	
		if ( ( start = time ( NULL ) ) == ( time_t ) -1 )
			ts_warning ( "main", 1 );
		input ( argv[1] );
		if ( ( end = time ( NULL ) ) == ( time_t ) -1 )
			ts_warning ( "main", 1 );

		read_time = difftime ( end, start );
		
		printf ( "%u vertices, %u triangles read\n", vs_size, ntr );
		if ( !ntr ) ts_error ( "main", 6 );
		
		/* create the array of sequential cycles, select the representative
		 * edge for each of the cycles, and create the Hopfield network */
		printf ( "Creating Hopfield network to \"hnet.txt\"\n" );
		
		if ( ( start = time ( NULL ) ) == ( time_t ) -1 )
			ts_warning ( "main", 1 );	
		hnet ( argv[1] );
		if ( ( end = time ( NULL ) ) == ( time_t ) -1 )
			ts_warning ( "main", 1 );

		create_time = difftime ( end, start );
            
		/* set the current minimum sufficiently high */
		mps = ntr + 1;
		
		/* set the number of trials to be performed */
		if ( argc < 4 )  nt = 10; else  nt = strtol ( argv[3], NULL, 10 );
		
		/* set the initial temperature */
		if ( argc < 5 )  t_init = 10.0; else  t_init = strtod ( argv[4], NULL );

		/* set the stopping criterion */
		if ( argc < 6 )  stop_crit = 0.3; else stop_crit = strtod ( argv[5], NULL );
		
		palf = pcas = ptau = 0.0;
		
		for ( i = 1; i <= nt; i++ ) {   
			printf ( "\n" );     
			printf ( "Trial no. %u:\n", i );
			printf ( "Computing by Hopfield net in \"tstr_tmp.txt\"\n" );

			/* set the value for the delay loop and start measuring the time of the computation */
			if ( ( st_delay = clock() ) == ( clock_t ) -1 )
				ts_warning ( "main", 1 );
			if ( ( start = time ( NULL ) ) == ( time_t ) -1 )
				ts_warning ( "main", 1 );
     
			/* run the computation */
			comp ( argv[1], &tt, t_init, stop_crit );

			/* stop measuring and calculate the running time */
			if ( ( end = time ( NULL ) ) == ( time_t ) -1 )
				ts_warning ( "main", 1 );
			comp_time = difftime ( end, start );
			
			printf ( "Running time (sec): %g\n", comp_time );
			printf ( "(%u iteration", tt );
			if ( tt != 1 )  printf ( "s" );
			printf ( ")\n" );

			/* generate temporary output file */
			printf ( "Generating sequential tristrips representation\n" );
			
			if ( trisgen ( argv[1], "tstr_tmp.objf", &ps, &tminlength, &tmaxlength ) ) { 
				/* here we get if the number of triangles read from the input file
				 * does not match the number of triangles generated by trisgen;
				 * this should occur only if there is a bug somewhere in the code */
				ts_error ( "main", 1 );
			}
			else {
				printf ( "Number of sequential tristrips: %u\n", ps );
				
				pcas += comp_time;
				ptau += (double) tt;
				palf += (double) ps;
				
				/* if we got a new minimum, save the trial and its statistics; 
				 * the trial is saved into the output file and into
				 * "comp.txt" by renaming the temporary files */
				
				/* temporary files exist */
				no_temp_files = 0;
				
				if (mps > ps) {
					tno = i;
					mps = ps;
					rtime = comp_time;
					iter = tt;
					minlength = tminlength;
					maxlength = tmaxlength;
					for ( j = 0; j < 10; j++ ) fcnt[j] = cnt[j];
					
					printf ( "Represents current minimum" );

					if ( remove ( argv[2] ) == -1 || remove ( "comp.txt" ) == -1 )
						ts_warning ( "main", 3 );
					
					if ( rename ( "tstr_tmp.objf", argv[2] ) == -1 ||
							rename ( "tstr_tmp.txt", "comp.txt" ) == -1 )
						ts_warning ( "main", 2 );
					else {
						printf ( " --> saved to \"%s\" and \"comp.txt\"\n", argv[2] );
					
						/* no need to remove temporary files */
						no_temp_files = 1;
					}
				}
			}

			/* a delaying loop */
			if ( st_delay != ( clock_t ) -1 )
				while ( st_delay == clock () ) ;
		}
		
		/* remove temporary files if necessary */
		if ( !no_temp_files ) {
			printf ( "\n" );
			printf ( "Removing temporary files\n" );
			if ( remove ( "tstr_tmp.objf" ) == -1 || remove ( "tstr_tmp.txt" ) == -1 )
				ts_warning ( "main", 3 );
		}
		
		printf ( "\n" );
		printf ( "======================================================================\n" );
		printf ( "\n" );
		printf ( "Input file \"%s\":\n", argv[1] );
		printf ( "\n" );
		printf ( "\t- number of vertices: %u\n", vs_size );
		printf ( "\t- number of triangles: %u\n", ntr );
		printf ( "\t- reading time (sec): %g\n", read_time );
		printf ( "\t- time needed for creating the Hopfield network (sec): %g\n", create_time );
		printf ( "\n" );
		printf ( "%u trial", nt );
		if ( nt > 1 ) printf ( "s" );
		printf ( " performed (initial temperature: %g, stopping criterion: %g):\n", t_init, stop_crit );
		printf ( "\n" );
		printf ( "\t- minimum number of sequential tristrips: %u\n", mps );
		printf ( "\t- average number of sequential tristrips: %g\n", palf / nt ); 
		printf ( "\t- average running time of the computation (sec): %g\n", pcas / nt ); 
		printf ( "\t- average convergence macroscopic time: %g\n", ptau / nt );    
		printf ( "\n" );
		printf ( "Statistics for the trial with minimum number of tristrips:\n" );
		printf ( "\n" );
		printf ( "\t- trial no.: %u\n", tno );
		printf ( "\t- running time (sec): %g\n", rtime );
		printf ( "\t- number of iterations: %u\n", iter );
		printf ( "\t- number of sequential tristrips: %u\n", mps );
		printf ( "\t- minimal tristrip length: %u\n", minlength );
		printf ( "\t- maximal tristrip length: %u\n", maxlength );
		printf ( "\t- average tristrip length: %g\n", (double) ntr / (double) mps );
		printf ( "\t- distribution of tristrip lengths (1,2,...,9,10 or more):\n" );
		printf ( "\t  \t[ " );
		for ( j = 0; j < 9; j++ ) printf ( "%u=%u ", j + 1, fcnt[j] );
		printf ( "10+=%u ]\n", fcnt[9] );
		printf ( "\n" );
		printf ( "Newly created files:\n" );
		printf ( "\n" );
		
		printf ( "\t- \"hnet.txt\" " );
		for ( j = 0; j < ( signed int ) ( strlen ( argv[2] ) - 8 ); j++ ) printf ( " " );
		printf ( "(Hopfield network details)\n" );
		
		printf ( "\t- \"comp.txt\" " );
		for ( j = 0; j < ( signed int ) ( strlen ( argv[2] ) - 8 ); j++ ) printf ( " " );
		printf ( "(optimal computation details)\n" );
		
		printf ( "\t- \"%s\" ", argv[2] );
		for ( j = 0; j < 8 - ( signed int ) strlen ( argv[2] ); j++ ) printf ( " " );
		printf ( "(resulting tristrip representation)\n" );
		printf ( "\n" );
	}
	return 0;
}
